package org.eclipse.jdt.internal.compiler.ast;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.ArrayAllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.ConditionalExpression;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.Invocation;
import org.eclipse.jdt.internal.compiler.ast.MessageSend;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.BinaryTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.CaptureBinding;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.Substitution;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBindingVisitor;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;

public class NullAnnotationMatching {
  public static final org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching NULL_ANNOTATIONS_OK = new org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching(Severity.OK, 1, null);
  
  public static final org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching NULL_ANNOTATIONS_OK_NONNULL = new org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching(Severity.OK, 4, null);
  
  public static final org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching NULL_ANNOTATIONS_UNCHECKED = new org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching(Severity.UNCHECKED, 1, null);
  
  public static final org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching NULL_ANNOTATIONS_MISMATCH = new org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching(Severity.MISMATCH, 1, null);
  
  private final Severity severity;
  
  public final TypeBinding superTypeHint;
  
  public final int nullStatus;
  
  NullAnnotationMatching(Severity severity, int nullStatus, TypeBinding superTypeHint) {
    this.severity = severity;
    this.superTypeHint = superTypeHint;
    this.nullStatus = nullStatus;
  }
  
  public boolean isAnyMismatch() {
    return this.severity.isAnyMismatch();
  }
  
  public boolean isUnchecked() {
    return (this.severity == Severity.UNCHECKED);
  }
  
  public boolean isDefiniteMismatch() {
    return (this.severity == Severity.MISMATCH);
  }
  
  public boolean wantToReport() {
    return (this.severity == Severity.LEGACY_WARNING);
  }
  
  public boolean isPotentiallyNullMismatch() {
    return (!isDefiniteMismatch() && this.nullStatus != -1 && (this.nullStatus & 0x10) != 0);
  }
  
  public String superTypeHintName(CompilerOptions options, boolean shortNames) {
    return String.valueOf(this.superTypeHint.nullAnnotatedReadableName(options, shortNames));
  }
  
  public static int checkAssignment(BlockScope currentScope, FlowContext flowContext, VariableBinding var, FlowInfo flowInfo, int nullStatus, Expression expression, TypeBinding providedType) {
    if (providedType == null)
      return 1; 
    long lhsTagBits = 0L;
    boolean hasReported = false;
    boolean usesNullTypeAnnotations = currentScope.environment().usesNullTypeAnnotations();
    if (!usesNullTypeAnnotations) {
      lhsTagBits = var.tagBits & 0x180000000000000L;
    } else {
      if (expression instanceof ConditionalExpression && expression.isPolyExpression()) {
        ConditionalExpression ce = (ConditionalExpression)expression;
        int status1 = checkAssignment(currentScope, flowContext, var, flowInfo, ce.ifTrueNullStatus, ce.valueIfTrue, ce.valueIfTrue.resolvedType);
        int status2 = checkAssignment(currentScope, flowContext, var, flowInfo, ce.ifFalseNullStatus, ce.valueIfFalse, ce.valueIfFalse.resolvedType);
        if (status1 == status2)
          return status1; 
        return nullStatus;
      } 
      lhsTagBits = var.type.tagBits & 0x180000000000000L;
      org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching annotationStatus = analyse(var.type, providedType, null, null, nullStatus, expression, CheckMode.COMPATIBLE);
      if (annotationStatus.isAnyMismatch()) {
        flowContext.recordNullityMismatch(currentScope, expression, providedType, var.type, flowInfo, nullStatus, annotationStatus);
        hasReported = true;
      } else {
        if (annotationStatus.wantToReport())
          annotationStatus.report((Scope)currentScope); 
        if (annotationStatus.nullStatus != 1)
          return annotationStatus.nullStatus; 
      } 
    } 
    if (lhsTagBits == 72057594037927936L && nullStatus != 4) {
      if (!hasReported)
        flowContext.recordNullityMismatch(currentScope, expression, providedType, var.type, flowInfo, nullStatus, null); 
      return 4;
    } 
    if (lhsTagBits == 36028797018963968L && nullStatus == 1) {
      if (usesNullTypeAnnotations && providedType.isTypeVariable() && (providedType.tagBits & 0x180000000000000L) == 0L)
        return 48; 
      return 24;
    } 
    return nullStatus;
  }
  
  public static org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching analyse(TypeBinding requiredType, TypeBinding providedType, int nullStatus) {
    return analyse(requiredType, providedType, null, null, nullStatus, null, CheckMode.COMPATIBLE);
  }
  
  public static org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching analyse(TypeBinding requiredType, TypeBinding providedType, TypeBinding providedSubstitute, Substitution substitution, int nullStatus, Expression providedExpression, CheckMode mode) {
    if (!requiredType.enterRecursiveFunction())
      return NULL_ANNOTATIONS_OK; 
    try {
      Severity severity = Severity.OK;
      TypeBinding superTypeHint = null;
      org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching okStatus = NULL_ANNOTATIONS_OK;
      if (areSameTypes(requiredType, providedType, providedSubstitute)) {
        if ((requiredType.tagBits & 0x100000000000000L) != 0L)
          return okNonNullStatus(providedExpression); 
        return okStatus;
      } 
      if (requiredType instanceof TypeVariableBinding && substitution != null && (mode == CheckMode.EXACT || mode == CheckMode.COMPATIBLE || mode == CheckMode.BOUND_SUPER_CHECK)) {
        requiredType.exitRecursiveFunction();
        requiredType = Scope.substitute(substitution, requiredType);
        if (!requiredType.enterRecursiveFunction())
          return NULL_ANNOTATIONS_OK; 
        if (areSameTypes(requiredType, providedType, providedSubstitute)) {
          if ((requiredType.tagBits & 0x100000000000000L) != 0L)
            return okNonNullStatus(providedExpression); 
          return okStatus;
        } 
      } 
      if (mode == CheckMode.BOUND_CHECK && requiredType instanceof TypeVariableBinding) {
        boolean passedBoundCheck = (substitution instanceof ParameterizedTypeBinding && (((ParameterizedTypeBinding)substitution).tagBits & 0x400000L) != 0L);
        if (!passedBoundCheck) {
          ReferenceBinding referenceBinding = requiredType.superclass();
          if (referenceBinding != null && (referenceBinding.hasNullTypeAnnotations() || substitution != null)) {
            org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching status = analyse((TypeBinding)referenceBinding, providedType, null, substitution, nullStatus, providedExpression, CheckMode.BOUND_SUPER_CHECK);
            severity = severity.max(status.severity);
            if (severity == Severity.MISMATCH)
              return new org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching(severity, nullStatus, superTypeHint); 
          } 
          ReferenceBinding[] arrayOfReferenceBinding = requiredType.superInterfaces();
          if (arrayOfReferenceBinding != null)
            for (int i = 0; i < arrayOfReferenceBinding.length; i++) {
              if (arrayOfReferenceBinding[i].hasNullTypeAnnotations() || substitution != null) {
                org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching status = analyse((TypeBinding)arrayOfReferenceBinding[i], providedType, null, substitution, nullStatus, providedExpression, CheckMode.BOUND_SUPER_CHECK);
                severity = severity.max(status.severity);
                if (severity == Severity.MISMATCH)
                  return new org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching(severity, nullStatus, superTypeHint); 
              } 
            }  
        } 
      } 
      if (requiredType instanceof ArrayBinding) {
        long[] requiredDimsTagBits = ((ArrayBinding)requiredType).nullTagBitsPerDimension;
        if (requiredDimsTagBits != null) {
          int dims = requiredType.dimensions();
          if (requiredType.dimensions() == providedType.dimensions()) {
            long[] providedDimsTagBits = ((ArrayBinding)providedType).nullTagBitsPerDimension;
            if (providedDimsTagBits == null)
              providedDimsTagBits = new long[dims + 1]; 
            int currentNullStatus = nullStatus;
            for (int i = 0; i <= dims; i++) {
              long requiredBits = validNullTagBits(requiredDimsTagBits[i]);
              long providedBits = validNullTagBits(providedDimsTagBits[i]);
              if (i == 0 && requiredBits == 36028797018963968L && nullStatus != -1 && mode.requiredNullableMatchesAll()) {
                if (nullStatus == 2)
                  break; 
              } else {
                if (i > 0)
                  currentNullStatus = -1; 
                Severity dimSeverity = computeNullProblemSeverity(requiredBits, providedBits, currentNullStatus, (i == 0) ? mode : mode.toDetail(), false);
                if (i > 0 && dimSeverity == Severity.UNCHECKED && 
                  providedExpression instanceof ArrayAllocationExpression && 
                  providedBits == 0L && requiredBits != 0L) {
                  Expression[] dimensions = ((ArrayAllocationExpression)providedExpression).dimensions;
                  Expression previousDim = dimensions[i - 1];
                  if (previousDim instanceof org.eclipse.jdt.internal.compiler.ast.IntLiteral && previousDim.constant.intValue() == 0) {
                    dimSeverity = Severity.OK;
                    nullStatus = -1;
                    break;
                  } 
                } 
                severity = severity.max(dimSeverity);
                if (severity == Severity.MISMATCH)
                  return NULL_ANNOTATIONS_MISMATCH; 
              } 
              if (severity == Severity.OK)
                nullStatus = -1; 
            } 
          } else if (providedType.id == 12 && 
            dims > 0 && requiredDimsTagBits[0] == 72057594037927936L) {
            return NULL_ANNOTATIONS_MISMATCH;
          } 
        } 
      } else if (requiredType.hasNullTypeAnnotations() || providedType.hasNullTypeAnnotations() || requiredType.isTypeVariable()) {
        long requiredBits = requiredNullTagBits(requiredType, mode);
        if (requiredBits != 36028797018963968L || nullStatus == -1 || !mode.requiredNullableMatchesAll()) {
          long providedBits = providedNullTagBits(providedType);
          Severity s = computeNullProblemSeverity(requiredBits, providedBits, nullStatus, mode, requiredType.isTypeVariable());
          if (s.isAnyMismatch() && requiredType.isWildcard() && requiredBits != 0L && (
            (WildcardBinding)requiredType).determineNullBitsFromDeclaration(null, null) == 0L)
            s = Severity.OK; 
          severity = severity.max(s);
          if (!severity.isAnyMismatch() && (providedBits & 0x100000000000000L) != 0L)
            okStatus = okNonNullStatus(providedExpression); 
        } 
        if (severity != Severity.MISMATCH && nullStatus != 2) {
          TypeBinding providedSuper = providedType.findSuperTypeOriginatingFrom(requiredType);
          TypeBinding providedSubstituteSuper = (providedSubstitute != null) ? providedSubstitute.findSuperTypeOriginatingFrom(requiredType) : null;
          if (severity == Severity.UNCHECKED && requiredType.isTypeVariable() && providedType.isTypeVariable() && (providedSuper == requiredType || providedSubstituteSuper == requiredType))
            severity = Severity.OK; 
          if (providedSuper != providedType)
            superTypeHint = providedSuper; 
          if (requiredType.isParameterizedType() && providedSuper instanceof ParameterizedTypeBinding) {
            TypeBinding[] requiredArguments = ((ParameterizedTypeBinding)requiredType).arguments;
            TypeBinding[] providedArguments = ((ParameterizedTypeBinding)providedSuper).arguments;
            TypeBinding[] providedSubstitutes = (providedSubstituteSuper instanceof ParameterizedTypeBinding) ? ((ParameterizedTypeBinding)providedSubstituteSuper).arguments : null;
            if (requiredArguments != null && providedArguments != null && requiredArguments.length == providedArguments.length)
              for (int i = 0; i < requiredArguments.length; i++) {
                TypeBinding providedArgSubstitute = (providedSubstitutes != null) ? providedSubstitutes[i] : null;
                org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching status = analyse(requiredArguments[i], providedArguments[i], providedArgSubstitute, substitution, -1, providedExpression, mode.toDetail());
                severity = severity.max(status.severity);
                if (severity == Severity.MISMATCH)
                  return new org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching(severity, nullStatus, superTypeHint); 
              }  
          } 
          ReferenceBinding referenceBinding1 = requiredType.enclosingType();
          ReferenceBinding referenceBinding2 = providedType.enclosingType();
          if (referenceBinding1 != null && referenceBinding2 != null) {
            ReferenceBinding referenceBinding = (providedSubstitute != null) ? providedSubstitute.enclosingType() : null;
            org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching status = analyse((TypeBinding)referenceBinding1, (TypeBinding)referenceBinding2, (TypeBinding)referenceBinding, substitution, -1, providedExpression, mode);
            severity = severity.max(status.severity);
          } 
        } 
      } 
      if (!severity.isAnyMismatch())
        return okStatus; 
      return new org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching(severity, nullStatus, superTypeHint);
    } finally {
      requiredType.exitRecursiveFunction();
    } 
  }
  
  public void report(Scope scope) {}
  
  public static org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching okNonNullStatus(Expression providedExpression) {
    if (providedExpression instanceof MessageSend) {
      MethodBinding method = ((MessageSend)providedExpression).binding;
      if (method != null && method.isValidBinding()) {
        MethodBinding originalMethod = method.original();
        ReferenceBinding referenceBinding = originalMethod.declaringClass;
        if (referenceBinding instanceof BinaryTypeBinding && 
          ((BinaryTypeBinding)referenceBinding).externalAnnotationStatus.isPotentiallyUnannotatedLib() && 
          originalMethod.returnType.isTypeVariable() && (
          originalMethod.returnType.tagBits & 0x180000000000000L) == 0L) {
          int severity = (((BinaryTypeBinding)referenceBinding).externalAnnotationStatus == BinaryTypeBinding.ExternalAnnotationStatus.NO_EEA_FILE) ? 
            0 : 1024;
          return (org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching)new Object(Severity.LEGACY_WARNING, 1, null, method, providedExpression, severity);
        } 
      } 
    } 
    return NULL_ANNOTATIONS_OK_NONNULL;
  }
  
  protected static boolean areSameTypes(TypeBinding requiredType, TypeBinding providedType, TypeBinding providedSubstitute) {
    if (requiredType == providedType)
      return true; 
    if (requiredType.isParameterizedType() || requiredType.isArrayType())
      return false; 
    if (TypeBinding.notEquals(requiredType, providedType)) {
      if (requiredType instanceof CaptureBinding) {
        TypeBinding lowerBound = ((CaptureBinding)requiredType).lowerBound;
        if (lowerBound != null && areSameTypes(lowerBound, providedType, providedSubstitute))
          return ((requiredType.tagBits & 0x180000000000000L) == (providedType.tagBits & 0x180000000000000L)); 
      } else {
        if (requiredType.kind() == 4100 && requiredType == providedSubstitute)
          return true; 
        if (providedType instanceof CaptureBinding) {
          TypeBinding upperBound = ((CaptureBinding)providedType).upperBound();
          if (upperBound != null && areSameTypes(requiredType, upperBound, providedSubstitute))
            return ((requiredType.tagBits & 0x180000000000000L) == (providedType.tagBits & 0x180000000000000L)); 
        } 
      } 
      return false;
    } 
    return ((requiredType.tagBits & 0x180000000000000L) == (providedType.tagBits & 0x180000000000000L));
  }
  
  static long requiredNullTagBits(TypeBinding type, CheckMode mode) {
    long tagBits = type.tagBits & 0x180000000000000L;
    if (tagBits != 0L)
      return validNullTagBits(tagBits); 
    if (type.isWildcard())
      return 108086391056891904L; 
    if (type.isTypeVariable()) {
      if (type.isCapture()) {
        TypeBinding lowerBound = ((CaptureBinding)type).lowerBound;
        if (lowerBound != null) {
          tagBits = lowerBound.tagBits & 0x180000000000000L;
          if (tagBits == 36028797018963968L)
            return 36028797018963968L; 
        } 
      } 
      switch (mode) {
        case null:
        case BOUND_SUPER_CHECK:
        case OVERRIDE_RETURN:
        case OVERRIDE:
          return 0L;
      } 
      return 72057594037927936L;
    } 
  }
  
  static long providedNullTagBits(TypeBinding type) {
    long tagBits = type.tagBits & 0x180000000000000L;
    if (tagBits != 0L)
      return validNullTagBits(tagBits); 
    if (type.isWildcard())
      return 108086391056891904L; 
    if (type.isTypeVariable()) {
      int i;
      TypeVariableBinding typeVariable = (TypeVariableBinding)type;
      boolean haveNullBits = false;
      if (typeVariable.isCapture()) {
        TypeBinding lowerBound = ((CaptureBinding)typeVariable).lowerBound;
        if (lowerBound != null) {
          tagBits = lowerBound.tagBits & 0x180000000000000L;
          if (tagBits == 36028797018963968L)
            return 36028797018963968L; 
          i = haveNullBits | ((tagBits != 0L) ? 1 : 0);
        } 
      } 
      if (typeVariable.firstBound != null) {
        long boundBits = typeVariable.firstBound.tagBits & 0x180000000000000L;
        if (boundBits == 72057594037927936L)
          return 72057594037927936L; 
        i |= (boundBits != 0L) ? 1 : 0;
      } 
      if (i != 0)
        return 108086391056891904L; 
    } 
    return 0L;
  }
  
  public static long validNullTagBits(long bits) {
    bits &= 0x180000000000000L;
    return (bits == 108086391056891904L) ? 0L : bits;
  }
  
  public static TypeBinding moreDangerousType(TypeBinding one, TypeBinding two) {
    if (one == null)
      return null; 
    long oneNullBits = validNullTagBits(one.tagBits);
    long twoNullBits = validNullTagBits(two.tagBits);
    if (oneNullBits != twoNullBits) {
      if (oneNullBits == 36028797018963968L)
        return one; 
      if (twoNullBits == 36028797018963968L)
        return two; 
      if (oneNullBits == 0L)
        return one; 
      return two;
    } 
    if (one != two && 
      analyse(one, two, -1).isAnyMismatch())
      return two; 
    return one;
  }
  
  private static Severity computeNullProblemSeverity(long requiredBits, long providedBits, int nullStatus, CheckMode mode, boolean requiredIsTypeVariable) {
    if (requiredBits == providedBits)
      return Severity.OK; 
    if (requiredBits == 0L) {
      switch (mode) {
        case COMPATIBLE:
        case EXACT:
        case null:
        case BOUND_SUPER_CHECK:
          return Severity.OK;
        case OVERRIDE_RETURN:
          if (providedBits == 72057594037927936L)
            return Severity.OK; 
          if (!requiredIsTypeVariable)
            return Severity.OK; 
          return Severity.UNCHECKED;
        case OVERRIDE:
          return Severity.UNCHECKED;
      } 
    } else {
      if (requiredBits == 108086391056891904L)
        return Severity.OK; 
      if (requiredBits == 72057594037927936L) {
        switch (mode) {
          case COMPATIBLE:
          case BOUND_SUPER_CHECK:
            if (nullStatus == 4)
              return Severity.OK; 
          case EXACT:
          case null:
          case OVERRIDE_RETURN:
          case OVERRIDE:
            if (providedBits == 0L)
              return Severity.UNCHECKED; 
            return Severity.MISMATCH;
        } 
      } else if (requiredBits == 36028797018963968L) {
        switch (mode) {
          case COMPATIBLE:
          case BOUND_SUPER_CHECK:
          case OVERRIDE_RETURN:
            return Severity.OK;
          case EXACT:
          case null:
            if (providedBits == 0L)
              return Severity.UNCHECKED; 
            return Severity.MISMATCH;
          case OVERRIDE:
            return Severity.MISMATCH;
        } 
      } 
    } 
    return Severity.OK;
  }
  
  public static MethodBinding checkForContradictions(MethodBinding method, Object location, Scope scope) {
    int start = 0, end = 0;
    if (location instanceof InvocationSite) {
      start = ((InvocationSite)location).sourceStart();
      end = ((InvocationSite)location).sourceEnd();
    } else if (location instanceof ASTNode) {
      start = ((ASTNode)location).sourceStart;
      end = ((ASTNode)location).sourceEnd;
    } 
    SearchContradictions searchContradiction = new SearchContradictions();
    TypeBindingVisitor.visit((TypeBindingVisitor)searchContradiction, method.returnType);
    if (searchContradiction.typeWithContradiction != null) {
      if (scope == null)
        return (MethodBinding)new ProblemMethodBinding(method, method.selector, method.parameters, 25); 
      scope.problemReporter().contradictoryNullAnnotationsInferred(method, start, end, location instanceof org.eclipse.jdt.internal.compiler.ast.FunctionalExpression);
      return method;
    } 
    Expression[] arguments = null;
    if (location instanceof Invocation)
      arguments = ((Invocation)location).arguments(); 
    for (int i = 0; i < method.parameters.length; i++) {
      TypeBindingVisitor.visit((TypeBindingVisitor)searchContradiction, method.parameters[i]);
      if (searchContradiction.typeWithContradiction != null) {
        if (scope == null)
          return (MethodBinding)new ProblemMethodBinding(method, method.selector, method.parameters, 25); 
        if (arguments != null && i < arguments.length) {
          scope.problemReporter().contradictoryNullAnnotationsInferred(method, (ASTNode)arguments[i]);
        } else {
          scope.problemReporter().contradictoryNullAnnotationsInferred(method, start, end, location instanceof org.eclipse.jdt.internal.compiler.ast.FunctionalExpression);
        } 
        return method;
      } 
    } 
    return method;
  }
  
  public static boolean hasContradictions(TypeBinding type) {
    SearchContradictions searchContradiction = new SearchContradictions();
    TypeBindingVisitor.visit((TypeBindingVisitor)searchContradiction, type);
    return (searchContradiction.typeWithContradiction != null);
  }
  
  public static TypeBinding strongerType(TypeBinding type1, TypeBinding type2, LookupEnvironment environment) {
    if ((type1.tagBits & 0x100000000000000L) != 0L)
      return mergeTypeAnnotations(type1, type2, true, environment); 
    return mergeTypeAnnotations(type2, type1, true, environment);
  }
  
  public static TypeBinding[] weakerTypes(TypeBinding[] parameters1, TypeBinding[] parameters2, LookupEnvironment environment) {
    TypeBinding[] newParameters = new TypeBinding[parameters1.length];
    for (int i = 0; i < newParameters.length; i++) {
      long tagBits1 = (parameters1[i]).tagBits;
      long tagBits2 = (parameters2[i]).tagBits;
      if ((tagBits1 & 0x80000000000000L) != 0L) {
        newParameters[i] = mergeTypeAnnotations(parameters1[i], parameters2[i], true, environment);
      } else if ((tagBits2 & 0x80000000000000L) != 0L) {
        newParameters[i] = mergeTypeAnnotations(parameters2[i], parameters1[i], true, environment);
      } else if ((tagBits1 & 0x100000000000000L) == 0L) {
        newParameters[i] = mergeTypeAnnotations(parameters1[i], parameters2[i], true, environment);
      } else {
        newParameters[i] = mergeTypeAnnotations(parameters2[i], parameters1[i], true, environment);
      } 
    } 
    return newParameters;
  }
  
  private static TypeBinding mergeTypeAnnotations(TypeBinding type, TypeBinding otherType, boolean top, LookupEnvironment environment) {
    TypeBinding mainType = type;
    if (!top) {
      AnnotationBinding[] otherAnnotations = otherType.getTypeAnnotations();
      if (otherAnnotations != Binding.NO_ANNOTATIONS)
        mainType = environment.createAnnotatedType(type, otherAnnotations); 
    } 
    if (mainType.isParameterizedType() && otherType.isParameterizedType()) {
      ParameterizedTypeBinding ptb = (ParameterizedTypeBinding)type, otherPTB = (ParameterizedTypeBinding)otherType;
      TypeBinding[] typeArguments = ptb.arguments;
      TypeBinding[] otherTypeArguments = otherPTB.arguments;
      TypeBinding[] newTypeArguments = new TypeBinding[typeArguments.length];
      for (int i = 0; i < typeArguments.length; i++)
        newTypeArguments[i] = mergeTypeAnnotations(typeArguments[i], otherTypeArguments[i], false, environment); 
      return (TypeBinding)environment.createParameterizedType(ptb.genericType(), newTypeArguments, ptb.enclosingType());
    } 
    return mainType;
  }
  
  public String toString() {
    if (this == NULL_ANNOTATIONS_OK)
      return "OK"; 
    if (this == NULL_ANNOTATIONS_MISMATCH)
      return "MISMATCH"; 
    if (this == NULL_ANNOTATIONS_OK_NONNULL)
      return "OK NonNull"; 
    if (this == NULL_ANNOTATIONS_UNCHECKED)
      return "UNCHECKED"; 
    StringBuilder buf = new StringBuilder();
    buf.append("Analysis result: severity=" + this.severity);
    buf.append(" nullStatus=" + this.nullStatus);
    return buf.toString();
  }
}
